unit ObjectsEgCarClasses;

interface

uses
  SysUtils;

type
  TCar = class(TObject)
  private
    FSpeed: Integer;
    FParked: Boolean;
    FTopSpeed: Integer;
    procedure SetParked(Value: Boolean);
    procedure SetSpeed(Value: Integer);
  protected
    procedure SetTopSpeed(Value: Integer); virtual;
  public
    constructor Create;
    procedure Drive; virtual;
    property Parked: Boolean read FParked write SetParked;
    property Speed: Integer read FSpeed write SetSpeed;
    property TopSpeed: Integer read FTopSpeed;
  end;

  TPorsche911 = class(TCar)
  public
    constructor Create;
    procedure Drive; override;
  end;

  TMorrisMinor = class(TCar)
  public
    constructor Create;
    procedure Drive; override;
  end;

  ECarError = class(Exception);

implementation

uses
  QDialogs;

procedure CarError(const ErrorMsg: String);
begin
  raise ECarError.Create(ErrorMsg)
end;

{ TCar }

constructor TCar.Create;
begin
  inherited Create;
  //Descendant car classes will set their top speeds
  SetTopSpeed(0);
  Parked := True;
end;

procedure TCar.Drive;
begin
  ShowMessage('Car is driving. Speed is %d', [FSpeed])
end;

procedure TCar.SetParked(Value: Boolean);
begin
  if Value <> FParked then
  begin
    if Value and (FSpeed > 0) then
      CarError('Car must be stationary to be parked');
    FParked := Value;
  end
end;

procedure TCar.SetSpeed(Value: Integer);
begin
  if Value <> FSpeed then
  begin
    if Value > TopSpeed then
      CarError(Format('Top speed of %d cannot be exceeded', [TopSpeed]));
    if Parked and (Value <> 0) then
      CarError('Cannot drive a parked car');
    FSpeed := Value
  end
end;

procedure TCar.SetTopSpeed(Value: Integer);
begin
  FTopSpeed := Value
end;

{ TPorsche911 }

constructor TPorsche911.Create;
begin
  inherited Create;
  SetTopSpeed(174); //top speed of some model of 911
end;

procedure TPorsche911.Drive;
begin
  ShowMessage('Driving in a Porsche 911 at %d mph (vroom, vroom!)', [Speed])
end;

{ TMorrisMinor }

constructor TMorrisMinor.Create;
begin
  inherited Create;
  SetTopSpeed(62); //top speed of some model of the MM
end;

procedure TMorrisMinor.Drive;
begin
  ShowMessage('Driving sedately in a Morris Minor at %d mph', [Speed])
end;

end.
